import json
import shutil
import subprocess
import sys
import tarfile
import vdf

from pathlib import Path
from os import environ
from urllib.error import HTTPError
from urllib.request import urlopen


WINETRICKS_URL = 'https://raw.githubusercontent.com/Winetricks/winetricks/master/src/winetricks'
DXVK_URL = "https://github.com/doitsujin/dxvk/releases/download/v{ver}/dxvk-{ver}.tar.gz"
DXVK_CACHE = Path("~/.cache/dxvk").expanduser()

DOTNET_REG = """
[Software\\Wow6432Node\\Valve\\Steam\\Apps\\CommonRedist\\.NET\\3.5] 1582602949
#time=1d5eb8f7748d36a
"3.5 SP1"=dword:00000001
"""


def get_path(path: str | Path) -> Path:
	"""
		Resolve a path

		:param path: Path to process
	"""

	if not isinstance(path, Path):
		path = Path(path)

	return path.expanduser().resolve()


class Config:
	"Base class for managing a WINE prefix for Project Torque"

	path: str = "~/.config/barkshark/pt-launcher.json"
	binpath: str = "~/.steam/steam/steamapps/common/Project Torque/ProjectTorque.exe"
	prefix: str = "~/.local/share/wineprefix/project-torque"
	winetricks: str = "~/.local/bin/winetricks"


	def __init__(self, binpath: str = None, prefix: str = None, winetricks: str = None):
		"""
			:param binpath: Path to the Project Torque executable
			:param prefix: Path to the WINE prefix
			:param winetricks: Path to the ``winetricks`` script
		"""

		self.path = get_path(Config.path)
		self.binpath = get_path(binpath or Config.binpath)
		self.prefix = get_path(prefix or Config.prefix)
		self.winetricks = get_path(winetricks or Config.winetricks)


	def fetch_dxvk(self, version: str):
		"""
			Download the specified version of DXVK if it hasn't been downloaded already

			:param version: Version of DXVK to fetch
			:raises ValueError: When an invalid version is specified
		"""

		DXVK_CACHE.mkdir(exist_ok = True, parents = True)

		dxvk_url = DXVK_URL.format(ver = version)
		dxvk_file = DXVK_CACHE.joinpath(f"dxvk-{version}.tar.gz")

		if not dxvk_file.exists():
			try:
				with urlopen(dxvk_url) as response, dxvk_file.open('wb') as out_file:
					shutil.copyfileobj(response, out_file)

			except HTTPError as e:
				if e.code == 404:
					raise ValueError(f"Invalid DXVK version: {version}") from None

				raise e

		if not DXVK_CACHE.joinpath(f"dxvk-{version}").exists():
			with tarfile.open(dxvk_file) as dxvk:
				dxvk.extractall(DXVK_CACHE)


	def fetch_winetricks(self):
		"Download winetricks"

		self.winetricks.parent.mkdir(exist_ok = True, parents = True)

		with urlopen(WINETRICKS_URL) as response, self.winetricks.open('wb') as out_file:
			shutil.copyfileobj(response, out_file)
			self.winetricks.chmod(0o755)


	def load(self):
		"Load the config file"

		with self.path.open("r", encoding = "utf-8") as fd:
			data = json.load(fd)

		for param in ("binpath", "prefix", "winetricks"):
			path = get_path(data.get(param, getattr(Config, param)))
			setattr(self, param, path)


	def locate_bin(self, directory: str = None) -> Path:
		"""
			Search for ``ProjectTorque.exe`` and set :attr:`Config.binpath` if it exists

			:param directory: Base directory to search. If one is not specified, the steam
				directory will be used instead.
			:raises FileNotFoundError: If the game binary cannot be found
		"""

		path = get_path(directory or "~/.steam")

		for item in path.rglob("*.exe"):
			if item.name.lower() == "projecttorque.exe":
				self.binpath = item
				return item

		raise FileNotFoundError("Cannot find 'ProjectTorque.exe'")


	def locate_bin_from_steam_config(self):
		"""
			Find the path to the Project Torque binary from the Steam config

			:raises FileNotFoundError: If the steam config path or the game bin could not be found
		"""

		paths = []
		steam_cfg_paths = [
			get_path("~/.steam/steam/config/libraryfolders.vdf"),
			get_path("~/.local/share/steam/config/libraryfolders.vdf")
		]

		for cfg_path in steam_cfg_paths:
			if not cfg_path.exists():
				continue

			with cfg_path.open("r", encoding = "UTF-8") as fd:
				cfg = vdf.load(fd, mapper = vdf.VDFDict)
				paths = list(item["path"] for item in cfg["libraryfolders"].values())

		if not paths:
			raise FileNotFoundError("Could not locate Steam library paths")

		for path in paths:
			try:
				self.locate_bin(path)
				return

			except FileNotFoundError:
				pass

		raise FileNotFoundError("Could not locate Project Torque from Steam config")


	def nuke_prefix(self):
		"Remove the WINE prefix"

		try:
			shutil.rmtree(self.prefix)

		except FileNotFoundError:
			pass


	def run(self) -> subprocess.CompletedProcess:
		"""
			Run Project Torque

			:raises FileNotFoundError: If the Project Torque binary cannot be found or the WINE
				prefix is not setup
		"""

		if not self.prefix.joinpath("setup").exists():
			raise FileNotFoundError("Wine prefix is not setup")

		if not self.binpath.exists():
			raise FileNotFoundError(f"Cannot find game binary at '{self.binpath}")

		# I tried using regedit to import a reg file, but it didn't work for some reason, so here's
		# a quick hack. This just prevents steam from trying to install dotnet3.5 every time it
		# starts.
		# if any(("common" in value) for value in sys.argv):
		# 	sysreg = Path(str(pt_path).lower().split('common')[0])
		# 	sysreg = sysreg..joinpath('compatdata/1112400/pfx/system.reg')
		# 	with sysreg.open('ab') as fd, script_path.joinpath('workaround.reg').open('rb') as reg:
		# 		fd.write(reg.read())

		return self.run_command("wine", self.binpath)


	def run_command(self, *command: str, cwd: str = None) -> subprocess.CompletedProcess:
		"""
			Run a command with extra environmental variables

			:param command: Command and arguments to run
			:param cwd: Use the specified directory instead of the Project Torque directory
		"""

		env = environ.copy()
		env["WINEPREFIX"] = self.prefix
		env["WINEARCH"] = "win32"
		env["W_OPT_UNATTENDED"] = "1"
		env["LD_PRELOAD"] = ""

		return subprocess.run(
			command,
			env = env,
			cwd = cwd or self.binpath.parent,
			check = False
		)


	def save(self):
		"Save the config to a file"

		params = ("binpath", "prefix", "winetricks")
		data = {param: getattr(self, param) for param in params}

		self.path.parent.mkdir(exist_ok = True, parents = True)

		with self.path.open("w", encoding = "utf-8") as fd:
			json.dump(data, fd, indent = "\t", default = str)


	def setup(self, dxvk_version: str = "2.3", nuke: bool = False):
		"""
			Setup the WINE prefix

			:param dxvk_version: Version of DXVK to use
			:param nuke: Run :meth:`Config.nuke_prefix` before setting up the prefix
		"""

		if self.prefix.joinpath("setup").exists() and not nuke:
			return

		if nuke:
			self.nuke_prefix()

		if not self.winetricks.exists():
			self.fetch_winetricks()

		winetricks_cmd = [
			self.winetricks,
			"dotnet35sp1",
			"tahoma",
			"d3dx9"

			# unneeded fixes?
			# "allfonts",
			# "d3dcompiler_43",
			# "d3dx9_36",
			# "d3dx9_41",
			# "mfc80"
		]

		if self.run_command(*winetricks_cmd).returncode != 0:
			raise RuntimeError("Winetricks returned invalid status code")

		self.setup_dxvk(dxvk_version)

		with self.prefix.joinpath("setup").open("w", encoding = "utf-8") as fd:
			fd.write("uwu")


	def setup_dxvk(self, version: str):
		"""
			Install DXVK into the WINE prefix

			:param version: Version of DXVK to use
		"""

		dxvk_dir = DXVK_CACHE.joinpath(f"dxvk-{version}/x32")

		self.fetch_dxvk(version)

		for dll in ("d3d9", "d3d10core", "d3d11", "dxgi"):
			shutil.copyfile(
				dxvk_dir.joinpath(f"{dll}.dll"),
				self.prefix.joinpath(f"drive_c/windows/system32/{dll}.dll")
			)

			proc = self.run_command(
				"wine", "reg", "add", "HKEY_CURRENT_USER\\Software\\Wine\\DllOverrides", "/v",
				dll, "/d", "native", "/f"
			)

			if proc.returncode != 0:
				raise RuntimeError(f"Failed to register dll: {dll}.dll")


def main():  # pylint: disable=missing-function-docstring
	config = Config()

	try:
		config.load()

	except FileNotFoundError:
		config.save()

	if sys.argv[-1].endswith("ProjectTorque.exe"):
		config.binpath = get_path(sys.argv[-1])

	config.setup()

	try:
		config.run()

	except KeyboardInterrupt:
		print('Bye')


if __name__ == "__main__":
	main()
